"""
Module to manage paths
"""
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Union

@contextmanager
def cd(path: Union[str, Path]):
    """
    Context manager to change the current working directory
    """
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)

def sanitize_path(path: str, allowed_dir: str) -> str:
    """
    Sanitizes a path, ensuring it's within the allowed directory
    """
    # Remove any trailing slashes
    path = os.path.normpath(path)

    # Replace backslashes with forward slashes
    path = path.replace("\\", "/")

    # Remove any relative path components
    path = os.path.abspath(path)

    # Check if the path is within the allowed directory
    allowed_dir = os.path.abspath(allowed_dir)
    if os.path.commonprefix([path, allowed_dir]) != allowed_dir:
        raise ValueError(f"Path {path} is not within allowed directory {allowed_dir}")

    return path



"""
Utilities for async functions
"""
import asyncio
from typing import Callable, Any, Coroutine, List, Dict
from functools import wraps


def make_sync(func: Callable[...,Coroutine[Any, Any, Any]]) -> Callable[...,Any]:
    """
    Decorator to run async functions synchronously
    Usage:

    @make_sync
    async def foo():
        pass

    """

    @wraps(func)
    def wrapper(*args: List[Any], **kwargs: Dict[Any, Any]) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return wrapper



"""
Perf measuring utilities
"""
import time
import os
import fcntl
from typing import Dict, Optional, List, Union
from itertools import chain


class _PP:
    def __init__(self, name: str):
        """
        name: name of the profiler
        """
        self.name: str = name
        self.tot_time: float = 0
        self.rounds: int = 0
        self.started: Optional[float] = None
        self.max_time = 0

    def start(self, allow_restart: bool = False) -> None:
        """
        Starts the perf counter
        """
        if not allow_restart and self.started is not None:
            raise RuntimeError("Already started")
        self.started = time.perf_counter()

    def stop(self, stop_time: float) -> None:
        """
        Stops the perf counter, trowhs an error if it was not started.
        Note: `stop_time` is the time returned by `time.perf_counter()`,
        this value is passed as parameter to avoid overhead.
        """
        if self.started is None:
            raise RuntimeError("Not started")
        runtime = stop_time - self.started
        self.max_time = max(self.max_time, runtime)
        self.tot_time += runtime
        self.rounds += 1
        self.started = None

    def stop_if_started(self, stop_time: float) -> None:
        """
        Stops the perf counter if it was started
        """
        if self.started is None:
            return
        self.stop(stop_time)


__pp__: Dict[str, _PP] = {}


def pp_start(name: str, allow_restart: bool = False):
    """
    Start a perf counter named `name`.
    If allow_restart is True, the perf counter can be started multiple times,
    in which case only the last start time is considered.
    """
    if name not in __pp__:
        __pp__[name] = _PP(name)
    # Start the perf counter as last as possible to avoid overhead
    __pp__[name].start(allow_restart)


def pp_stop(name: str, allow_restop: bool = False):
    """
    Stop a perf counter named `name`.
    If allow_restop is True, the perf counter can be stopped multiple times,
    in which case only the first stop time is considered.
    """
    # Stop the perf counter as soon as possible to avoid overhead
    stop_time = time.perf_counter()
    if name not in __pp__:
        __pp__[name] = _PP(name)
    if not allow_restop:
        __pp__[name].stop(stop_time)
    else:
        __pp__[name].stop_if_started(stop_time)

def pp_get(name: str) -> _PP:
    """
    Return the perf counter named `name`
    """
    return __pp__[name]

def pp_reset(name: Optional[str] = None) -> None:
    """
    Reset the perf counter named `name`.
    If `name` is None, reset all perf counters.
    """
    if name is None:
        __pp__.clear()
    else:
        del __pp__[name]

def pp_stats():
    """ 
    Pretty print perf counters stats
    """

    def _truncate(string: str, max_length: int):
        """
        Truncate a string to `max_length` characters,
        adding "..." if the string is longer than `max_length`
        """
        if len(string) > max_length:
            return string[: max_length - 3] + "..."
        return string

    def _print_row(row: Union[List[str], List[Union[str, int, float]]], column_widths: List[int]):
        """
        Print a row of the table
        """
        formatted_row = [_truncate(str(item), w).ljust(w) for item, w in zip(row, column_widths)]
        print(" | ".join(formatted_row))

    # Determine the maximum width of each column
    headers: List[str] = ["PID", "Name", "tot_time", "rounds", "avg_time", "max_time"]
    rows: List[List[Union[str, int, float]]] = [
        [os.getpid(), pp.name, pp.tot_time, pp.rounds, pp.tot_time / pp.rounds, pp.max_time]
        for pp in sorted(__pp__.values(), key=lambda pp: pp.tot_time, reverse=True)
    ]

    # Ensure that all rows have the same number of columns
    assert all(len(headers) == len(r) for r in rows)

    # Determine the maximum width of each column
    max_col_width = 25
    column_widths = [
        min(max_col_width, max(len(str(r[i])) for r in chain([headers], rows)))
        for i in range(len(headers))
    ]

    # Print the table
    with open(os.path.expanduser("~/.frittomisto_lock"), "w", encoding="utf-8") as f:
        fcntl.flock(f, fcntl.LOCK_EX)
        try:
            _print_row(headers, column_widths)
            for pp in sorted(__pp__.values(), key=lambda pp: pp.tot_time, reverse=True):
                _print_row(
                    [
                        os.getpid(),
                        pp.name,
                        pp.tot_time,
                        pp.rounds,
                        pp.tot_time / pp.rounds,
                        pp.max_time,
                    ],
                    column_widths,
                )
            print()
        finally:
            fcntl.flock(f, fcntl.LOCK_UN)



"""
Module to manage config files
"""
from typing import Any, Dict, Union, List, Optional
from pathlib import Path
import toml


class FrittoMistoCfg:
    """
    Class to manage the frittomisto config files.
    The config file is a TOML file, and the default name is frittomisto.toml and it
    is searched for in the current directory and its parents.

    The API of this class is similar to the one of python Dicts:
    - cfg['key'] has a similar effect to dict['key']
    - cfg.get('key', default) has a similar effect to dict.get('key', default)

    It is possible to modify the behaviour of this class by calling the
    "set_" methods. For example, to set the config file to use, call
    set_config_file('path/to/config.toml').
    """

    def __init__(self):
        self._config: Optional[Dict[str, Any]] = None
        self._config_names: List[str] = ["frittomisto.toml"]
        self._config_file: Optional[str] = None

    def set_config_file(self, config_file: Optional[str]) -> None:
        """
        Set the config file to use
        """
        self._config_file = config_file

    def set_config_names(self, config_names: Union[str, List[str]]) -> None:
        """
        Set the config names to search for, in order of priority.
        The default is ['frittomisto.toml']
        """
        if not config_names:
            raise ValueError("config_names must be a non-empty string or list")

        if isinstance(config_names, str):
            config_names = [config_names]

        self._config_names = config_names

    def get_config_names(self) -> List[str]:
        """
        Get the config names to search for, in order of priority.
        The default is ['frittomisto.toml']
        """
        return self._config_names

    def reload(self) -> None:
        """
        Reload the config from the config file
        """
        self._config = self._reload_config()

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Get a config value
        """
        return self._get_config().get(key, default)

    def __getitem__(self, __name: str) -> Any:
        """
        Get a config value
        """
        return self._get_config()[__name]

    def __setitem__(self, __name: str, __value: Any) -> None:
        raise NotImplementedError("Setting config values is not yet implemented")

    def _get_config(self) -> Dict[str, Any]:
        # Find the config file if it hasn't been found yet
        if self._config_file is None:
            self._config_file = self._find_config()

        if self._config is None:
            self._config = self._reload_config()

        return self._config

    def _reload_config(self) -> Dict[str, Any]:
        """
        Reload the config from the config file
        """
        if self._config_file is None:
            raise ValueError("No config file set")

        with open(self._config_file, encoding="utf-8") as f:
            config = toml.load(f)
        return config

    def _find_config(self) -> str:
        """
        Find config file in the current directory or its parents
        """
        for directory in [Path("."), *Path(".").absolute().parents]:
            for name in self._config_names:
                if (directory / name).exists():
                    return str(directory / name)

        raise FileNotFoundError("Could not find config file in current directory or its parents")


cfg = FrittoMistoCfg()



"""
Utilities for logging
"""
import logging
from typing import Optional, Generator, Union
from contextlib import contextmanager
from functools import lru_cache

@lru_cache(maxsize=None) # Cache the result to avoid registering the same handler multiple times
def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Returns a logger with the given name.
    Set the level of the logger with logger.setLevel(logging.DEBUG)
    """
    if name is None:
        logger = logging.getLogger()
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    else:
        logger = logging.getLogger(name)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.propagate = False
    return logger

@contextmanager
def log_level(level: Union[int, str], name: Optional[str] = None) -> Generator[None, None, None]:
    """
    Context manager to set the log level for the given logger
    """
    logger = logging.getLogger(name)
    original_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(original_level)



"""
Module for managing outputs
"""
from unittest.mock import patch
from contextlib import contextmanager

@contextmanager
def no_print():
    """
    Context manager to disable printing, use it like this:
    with no_print():
        print("This will not be printed")
        func() # This will not print anything either
    """
    with patch("builtins.print") as mock_print:
        mock_print.side_effect = lambda : None
        yield
